﻿
using System;
using System.Collections.Generic;
using System.IO;
using System.Text;

namespace etoy
{
    interface IJsonWritable 
    {
        /// <summary>
        /// 迭代器写入
        /// 会自动拼接 ','
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="enumerable"></param>
        /// <param name="len"></param>
        /// <param name="callback"></param>
        /// <returns></returns>
        IJsonWritable EnumeratorToJson<T>(IEnumerable<T> enumerable, int len, Action<int, T> callback);

        /// <summary>
        /// 写入key 自动拼接成 -> \"key\":
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        IJsonWritable AppendKey(string value);

        /// <summary>
        /// 写入value
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        IJsonWritable Append(int value);

        /// <summary>
        /// 写入value -> \"value\"
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        IJsonWritable Append(string value);

        /// <summary>
        /// 强制写入value 不会进行转义
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        IJsonWritable AppendForce(string value);

        /// <summary>
        /// [
        /// </summary>
        /// <returns></returns>
        IJsonWritable ArrayBracket();

        /// <summary>
        /// ]
        /// </summary>
        /// <returns></returns>
        IJsonWritable EndArrayBracket();

        /// <summary>
        /// {
        /// </summary>
        /// <returns></returns>
        IJsonWritable Bracket();

        /// <summary>
        /// }
        /// </summary>
        /// <returns></returns>
        IJsonWritable EndBracket();

        /// <summary>
        /// ,
        /// </summary>
        /// <returns></returns>
        IJsonWritable Dot();

        /// <summary>
        /// 换行
        /// </summary>
        /// <returns></returns>
        IJsonWritable NewLine();

        IJsonWritable Indent();

        IJsonWritable Dedent();
    }

    class JsonWriter : IDisposable
        , IJsonWritable
    {
        #region Constructors


        private readonly TextWriter _w;
        private MemoryStream _ms = new MemoryStream();
        private readonly bool _formatting;
        private Encoding _utf8WithoutBom = new UTF8Encoding(false);

        public JsonWriter(bool formatting = false) 
        {
            _ms = new MemoryStream();
            _w = new StreamWriter(_ms, _utf8WithoutBom);
            _formatting = formatting;
        }

        /// <summary>
        /// Return the generated code as a string.
        /// </summary>
        public string Json
        {
            get
            {
                _w.Flush();
                return _utf8WithoutBom.GetString(_ms.ToArray());
            }
        }

        public virtual void Dispose()
        {
            _w.Close();
        }
        #endregion

        #region Json Write

        public IJsonWritable EnumeratorToJson<T>(IEnumerable<T> enumerable, int len, Action<int, T> callback)
        {
            var count = 0;
            var enumerator = enumerable.GetEnumerator();
            while (enumerator.MoveNext())
            {
                callback.Invoke(count, enumerator.Current);
                if (++count >= len) break;
                this.Dot();
            }
            enumerator.Dispose();
            return this;
        }

        public IJsonWritable AppendKey(string value)
        {
            _w.Write($"\"{value}\":");
            return this;
        }

        public IJsonWritable Append(int value)
        {
            _w.Write(value);
            return this;
        }

        public IJsonWritable Append(string value)
        {
            _w.Write($"\"{value}\"");
            return this;
        }

        /// <summary>
        /// 不拼接引号强制写入
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public IJsonWritable AppendForce(string value)
        {
            _w.Write($"{value}");
            return this;
        }

        public IJsonWritable ArrayBracket()
        {
            _w.Write('[');
            return this;
        }
        public IJsonWritable EndArrayBracket()
        {
            _w.Write(']');
            return this;
        }

        public IJsonWritable Bracket()
        {
            _w.Write('{');
            return this;
        }

        public IJsonWritable Dot()
        {
            _w.Write(',');
            return this;
        }

        public IJsonWritable EndBracket()
        {
            _w.Write('}');
            return this;
        }
        #endregion


        #region 格式化

        public static string IndentPrefix = "    ";

        /// <summary>
        /// Level of indentation
        /// </summary>
        public int IndentLevel { get; private set; }

        /// <summary>
        /// Accumulated prefixes over indentations
        /// </summary>
        string prefix = "";

        ///<summary> 增加缩进 </summary>
        public IJsonWritable Indent()
        {
            IndentLevel += 1;
            prefix += IndentPrefix;
            return this;
        }

        ///<summary> 减少缩进 </summary>
        public IJsonWritable Dedent()
        {
            IndentLevel -= 1;
            if (IndentLevel < 0)
                throw new InvalidOperationException("Indent error");
            prefix = prefix.Substring(0, prefix.Length - IndentPrefix.Length);
            return this;
        }

        public IJsonWritable NewLine()
        {
            if (_formatting) 
            {
                _w.Write("\r\n");
                _w.Write(prefix);
            }
            return this;
        }

        #endregion
    }
}
